from typing import Any, Dict, Optional, List
from propelauth_py.types.user import UserMetadata, OrgFromApiKey
from propelauth_py.user import OrgMemberInfo
from dataclasses import dataclass

@dataclass
class ApiKeyFull:
    api_key_id: str
    created_at: int
    expires_at_seconds: Optional[int]
    metadata: Optional[Dict[str, Any]]
    user_id: Optional[str]
    org_id: Optional[str]
    
    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class ApiKeyResultPage:
    api_keys: List[ApiKeyFull]
    total_api_keys: int
    current_page: int
    page_size: int
    has_more_results: bool

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class ApiKeyNew:
    api_key_id: str
    api_key_token: str

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class ApiKeyValidation:
    metadata: Optional[Dict[str, Any]]
    user: Optional[UserMetadata]
    org: Optional[OrgFromApiKey]
    user_in_org: Optional[OrgMemberInfo]

    def __getitem__(self, key):
        return getattr(self, key)



from propelauth_py.errors import UnauthorizedException, ForbiddenException
from propelauth_py.jwt import _validate_access_token_and_get_user
from propelauth_py.user import UserAndOrgMemberInfo

# wrapper functions

def wrap_validate_access_token_and_get_user(token_verification_metadata):
    def validate_access_token_and_get_user(authorization_header):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, token_verification_metadata)
        return user

    return validate_access_token_and_get_user


def wrap_validate_access_token_and_get_user_with_org(token_verification_metadata):
    def validate_access_token_and_get_user_with_org(authorization_header, required_org_id):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, token_verification_metadata)
        org_member_info = validate_org_access_and_get_org_member_info(user, required_org_id)
        return UserAndOrgMemberInfo(user, org_member_info)

    return validate_access_token_and_get_user_with_org


def wrap_validate_access_token_and_get_user_with_org_by_minimum_role(token_verification_metadata):
    def validate_access_token_and_get_user_with_org_by_minimum_role(authorization_header, required_org_id, minimum_required_role):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, token_verification_metadata)
        org_member_info = validate_minimum_org_role_and_get_org(user, required_org_id, minimum_required_role)
        return UserAndOrgMemberInfo(user, org_member_info)

    return validate_access_token_and_get_user_with_org_by_minimum_role


def wrap_validate_access_token_and_get_user_with_org_by_exact_role(token_verification_metadata):
    def validate_access_token_and_get_user_with_org_by_exact_role(authorization_header, required_org_id, required_role):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, token_verification_metadata)
        org_member_info = validate_exact_org_role_and_get_org(user, required_org_id, required_role)
        return UserAndOrgMemberInfo(user, org_member_info)

    return validate_access_token_and_get_user_with_org_by_exact_role


def wrap_validate_access_token_and_get_user_with_org_by_permission(token_verification_metadata):
    def validate_access_token_and_get_user_with_org_by_permission(authorization_header, required_org_id, permission):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, token_verification_metadata)
        org_member_info = validate_permission_and_get_org(user, required_org_id, permission)
        return UserAndOrgMemberInfo(user, org_member_info)

    return validate_access_token_and_get_user_with_org_by_permission


def wrap_validate_access_token_and_get_user_with_org_by_all_permissions(token_verification_metadata):
    def validate_access_token_and_get_user_with_org_by_all_permissions(authorization_header, required_org_id, permissions):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, token_verification_metadata)
        org_member_info = validate_all_permissions_and_get_org(user, required_org_id, permissions)
        return UserAndOrgMemberInfo(user, org_member_info)

    return validate_access_token_and_get_user_with_org_by_all_permissions

# validation functions


def validate_org_access_and_get_org_member_info(user, required_org_id):
    """performs basic verifications and returns the org member info"""
    if required_org_id is None:
        raise ForbiddenException.unknown_required_org()

    org_id_to_org_member_info = user.org_id_to_org_member_info
    if org_id_to_org_member_info is None:
        raise ForbiddenException.user_not_member_of_org(required_org_id)

    org_member_info = org_id_to_org_member_info.get(required_org_id)
    if org_member_info is None:
        raise ForbiddenException.user_not_member_of_org(required_org_id)

    return org_member_info


def validate_minimum_org_role_and_get_org(user, required_org_id, minimum_role):
    """returns the org is the user is or inherits from the role specified"""
    org_member_info = validate_org_access_and_get_org_member_info(user, required_org_id)

    if minimum_role is not None and not org_member_info.user_is_at_least_role(minimum_role):
        raise ForbiddenException.user_doesnt_have_required_role()

    return org_member_info


def validate_exact_org_role_and_get_org(user, required_org_id, exact_role):
    """returns the org is the user has the exact role specified"""
    org_member_info = validate_org_access_and_get_org_member_info(user, required_org_id)

    if exact_role is not None and not org_member_info.user_is_role(exact_role):
        raise ForbiddenException.user_doesnt_have_required_role()

    return org_member_info


def validate_permission_and_get_org(user, required_org_id, permission):
    """returns the org is the user has the permission specified"""
    org_member_info = validate_org_access_and_get_org_member_info(user, required_org_id)

    if permission is not None and not org_member_info.user_has_permission(permission):
        raise ForbiddenException.user_doesnt_have_required_permission()

    return org_member_info


def validate_all_permissions_and_get_org(user, required_org_id, permissions):
    """returns the org is the user has the permission specified"""
    org_member_info = validate_org_access_and_get_org_member_info(user, required_org_id)

    if permissions is not None and not org_member_info.user_has_all_permissions(permissions):
        raise ForbiddenException.user_doesnt_have_required_permission()

    return org_member_info

# helper functions

def _extract_token_from_authorization_header(authorization_header):
    if authorization_header is None or authorization_header == "":
        raise UnauthorizedException.no_header_found()

    auth_header_parts = authorization_header.split(" ")
    if len(auth_header_parts) != 2 or auth_header_parts[0].lower() != "bearer":
        raise UnauthorizedException.invalid_header_found()

    return auth_header_parts[1]



import requests
from propelauth_py.api import _ApiKeyAuth, _is_valid_id, _auth_hostname_header, BACKEND_API_BASE_URL
from propelauth_py.errors import BadRequestException, UserNotFoundException, RateLimitedException

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/backend/v1/access_token"

class CreateAccessTokenResponse:
    def __init__(
        self,
        access_token: str,
    ):
        self.access_token = access_token

    def __repr__(self): 
        return (
            f"CreateAccessTokenResponse(access_token={self.access_token}"
        )
    def __eq__(self, other):
        return isinstance(other, CreateAccessTokenResponse)

def _create_access_token(auth_hostname, integration_api_key, user_id, duration_in_minutes) -> CreateAccessTokenResponse:
    if not _is_valid_id(user_id):
        raise UserNotFoundException()

    url = ENDPOINT_URL
    json = {"user_id": user_id, "duration_in_minutes": duration_in_minutes}
    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 403:
        raise UserNotFoundException()
    elif response.status_code == 404:
        raise RuntimeError("Access token creation is not enabled")
    elif not response.ok:
        raise RuntimeError("Unknown error when creating access token")

    json_response = response.json()
    return CreateAccessTokenResponse(
        access_token=json_response.get('access_token')
    )



from typing import Optional
import requests
from propelauth_py.api import (
    _ApiKeyAuth,
    _is_valid_hex,
    remove_bearer_if_exists,
    BACKEND_API_BASE_URL,
)
from propelauth_py.errors import (
    EndUserApiKeyException,
    EndUserApiKeyNotFoundException,
    EndUserApiKeyRateLimitedException,
    RateLimitedException,
)
from propelauth_py.types.end_user_api_keys import (
    ApiKeyFull,
    ApiKeyResultPage,
    ApiKeyNew,
    ApiKeyValidation,
)
from propelauth_py.types.user import UserMetadata, OrgFromApiKey
from propelauth_py.user import OrgMemberInfo

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/backend/v1/end_user_api_keys"


####################
#       GET        #
####################
def _fetch_api_key(auth_hostname, integration_api_key, api_key_id) -> ApiKeyFull:
    if not _is_valid_hex(api_key_id):
        raise EndUserApiKeyNotFoundException()

    url = f"{ENDPOINT_URL}/{api_key_id}"    
    response = requests.get(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching end user api key")

    json_response = response.json()
    return ApiKeyFull(
        api_key_id=json_response.get("api_key_id"),
        created_at=json_response.get("created_at"),
        expires_at_seconds=json_response.get("expires_at_seconds"),
        metadata=json_response.get("metadata"),
        user_id=json_response.get("user_id"),
        org_id=json_response.get("org_id"),
    )


def _fetch_current_api_keys(
    auth_hostname,
    integration_api_key,
    org_id=None,
    user_id=None,
    user_email=None,
    page_size=None,
    page_number=None,
    api_key_type=None,
) -> ApiKeyResultPage:
    url = ENDPOINT_URL

    query_params = {}
    if org_id:
        query_params["org_id"] = org_id
    if user_id:
        query_params["user_id"] = user_id
    if user_email:
        query_params["user_email"] = user_email
    if page_size:
        query_params["page_size"] = page_size
    if page_number:
        query_params["page_number"] = page_number
    if api_key_type:
        query_params["api_key_type"] = api_key_type

    response = requests.get(
        url, 
        auth=_ApiKeyAuth(integration_api_key),
        params=query_params,
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching current end user api keys")

    json_response = response.json()

    api_keys = [
        ApiKeyFull(
            api_key_id=key.get("api_key_id"),
            created_at=key.get("created_at"),
            expires_at_seconds=key.get("expires_at_seconds"),
            metadata=key.get("metadata"),
            user_id=key.get("user_id"),
            org_id=key.get("org_id"),
        )
        for key in json_response.get("api_keys")
    ]

    return ApiKeyResultPage(
        api_keys=api_keys,
        total_api_keys=json_response.get("total_api_keys"),
        current_page=json_response.get("current_page"),
        page_size=json_response.get("page_size"),
        has_more_results=json_response.get("has_more_results"),
    )


def _fetch_archived_api_keys(
    auth_hostname,
    integration_api_key,
    org_id=None,
    user_id=None,
    user_email=None,
    page_size=None,
    page_number=None,
    api_key_type=None,
) -> ApiKeyResultPage:
    url = f"{ENDPOINT_URL}/archived"

    query_params = {}
    if org_id:
        query_params["org_id"] = org_id
    if user_id:
        query_params["user_id"] = user_id
    if user_email:
        query_params["user_email"] = user_email
    if page_size:
        query_params["page_size"] = page_size
    if page_number:
        query_params["page_number"] = page_number
    if api_key_type:
        query_params["api_key_type"] = api_key_type

    response = requests.get(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        params=query_params,
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching archived end user api keys")

    json_response = response.json()

    api_keys = [
        ApiKeyFull(
            api_key_id=key.get("api_key_id"),
            created_at=key.get("created_at"),
            expires_at_seconds=key.get("expires_at_seconds"),
            metadata=key.get("metadata"),
            user_id=key.get("user_id"),
            org_id=key.get("org_id"),
        )
        for key in json_response.get("api_keys")
    ]

    return ApiKeyResultPage(
        api_keys=api_keys,
        total_api_keys=json_response.get("total_api_keys"),
        current_page=json_response.get("current_page"),
        page_size=json_response.get("page_size"),
        has_more_results=json_response.get("has_more_results"),
    )


####################
#       POST       #
####################
def _create_api_key(
    auth_hostname, integration_api_key, org_id, user_id, expires_at_seconds, metadata
) -> ApiKeyNew:
    url = ENDPOINT_URL

    json = {}
    if org_id:
        json["org_id"] = org_id
    if user_id:
        json["user_id"] = user_id
    if expires_at_seconds:
        json["expires_at_seconds"] = expires_at_seconds
    if metadata:
        json["metadata"] = metadata

    response = requests.post(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        json=json,
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when creating end user api key")

    json_response = response.json()
    return ApiKeyNew(
        api_key_id=json_response.get("api_key_id"),
        api_key_token=json_response.get("api_key_token"),
    )


def _validate_api_key(auth_hostname, integration_api_key, api_key_token) -> ApiKeyValidation:
    url = f"{ENDPOINT_URL}/validate"
    json = {"api_key_token": remove_bearer_if_exists(api_key_token)}
    response = requests.post(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        json=json,
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif response.status_code == 429:
        try:
            end_user_rate_limit_response = response.json()
            raise EndUserApiKeyRateLimitedException(end_user_rate_limit_response)
        except requests.exceptions.JSONDecodeError:
            raise RateLimitedException(response.text)
    elif not response.ok:
        raise RuntimeError("Unknown error when validating end user api key")

    json_response = response.json()

    user = None
    if json_response.get("user") is not None:
        user_data = json_response.get("user")
        user = UserMetadata(
            user_id=user_data.get("user_id"),
            email=user_data.get("email"),
            email_confirmed=user_data.get("email_confirmed"),
            has_password=user_data.get("has_password"),
            username=user_data.get("username"),
            first_name=user_data.get("first_name"),
            last_name=user_data.get("last_name"),
            picture_url=user_data.get("picture_url"),
            locked=user_data.get("locked"),
            enabled=user_data.get("enabled"),
            mfa_enabled=user_data.get("mfa_enabled"),
            can_create_orgs=user_data.get("can_create_orgs"),
            created_at=user_data.get("created_at"),
            last_active_at=user_data.get("last_active_at"),
            org_id_to_org_info=user_data.get("org_id_to_org_info"),
            legacy_org_id=user_data.get("legacy_org_id"),
            impersonator_user_id=user_data.get("impersonator_user_id"),
            metadata=user_data.get("metadata"),
            properties=user_data.get("properties"),
        )

    org = None
    if json_response.get("org") is not None:
        org_data = json_response.get("org")
        org = OrgFromApiKey(
            org_id=org_data.get("org_id"),
            name=org_data.get("org_name"),
            org_name=org_data.get("org_name"),
            max_users=org_data.get("max_users"),
            is_saml_configured=org_data.get("is_saml_configured"),
            legacy_org_id=org_data.get("legacy_org_id"),
            metadata=org_data.get("metadata", {}),
            custom_role_mapping_name=org_data.get("custom_role_mapping_name"),
        )

    user_in_org = None
    if json_response.get("user_in_org") is not None:
        user_in_org_data = json_response.get("user_in_org")
        user_in_org = OrgMemberInfo(
            org_id=user_in_org_data.get("org_id"),
            org_name=user_in_org_data.get("org_name"),
            org_metadata=user_in_org_data.get("org_metadata"),
            user_assigned_role=user_in_org_data.get("user_role"),
            url_safe_org_name=user_in_org_data.get("url_safe_org_name"),
            user_inherited_roles_plus_current_role=user_in_org_data.get(
                "inherited_user_roles_plus_current_role"
            ),
            user_permissions=user_in_org_data.get("user_permissions"),
            org_role_structure=user_in_org_data.get("org_role_structure"),
            assigned_additional_roles=user_in_org_data.get("additional_roles", []),
            legacy_org_id=user_in_org_data.get("legacy_org_id"),
        )

    return ApiKeyValidation(
        metadata=json_response.get("metadata"),
        user=user,
        org=org,
        user_in_org=user_in_org,
    )


####################
#    PUT/PATCH     #
####################
def _update_api_key(
    auth_hostname, integration_api_key, api_key_id, expires_at_seconds, metadata
) -> bool:
    if not _is_valid_hex(api_key_id):
        return False

    url = f"{ENDPOINT_URL}/{api_key_id}"

    json = {}
    if expires_at_seconds:
        json["expires_at_seconds"] = expires_at_seconds
    if metadata:
        json["metadata"] = metadata

    response = requests.patch(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        json=json,
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif not response.ok:
        raise RuntimeError("Unknown error when updating end user api key")

    return True


####################
#      DELETE      #
####################
def _delete_api_key(auth_hostname, integration_api_key, api_key_id) -> bool:
    if not _is_valid_hex(api_key_id):
        return False

    url = f"{ENDPOINT_URL}/{api_key_id}"
    response = requests.delete(
        url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif not response.ok:
        raise RuntimeError("Unknown error when deleting end user api key")

    return True



import requests

from propelauth_py.api import _ApiKeyAuth, _auth_hostname_header, BACKEND_API_BASE_URL
from propelauth_py.errors import BadRequestException, RateLimitedException
from propelauth_py.types.user import CreatedUser

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/backend/v1/migrate_user"


def _migrate_user_from_external_source(
    auth_hostname,
    integration_api_key,
    email,
    email_confirmed,
    existing_user_id=None,
    existing_password_hash=None,
    existing_mfa_base32_encoded_secret=None,
    ask_user_to_update_password_on_login=False,
    enabled=None,
    first_name=None,
    last_name=None,
    username=None,
    picture_url=None,
    properties=None,
) -> CreatedUser:
    url = ENDPOINT_URL + "/"
    json = {
        "email": email,
        "email_confirmed": email_confirmed,
        "existing_user_id": existing_user_id,
        "existing_password_hash": existing_password_hash,
        "existing_mfa_base32_encoded_secret": existing_mfa_base32_encoded_secret,
        "update_password_required": ask_user_to_update_password_on_login,
        "enabled": enabled,
        "first_name": first_name,
        "last_name": last_name,
        "username": username,
        "picture_url": picture_url,
    }
    if properties is not None:
        json["properties"] = properties

    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when migrating user")

    json_response = response.json()
    return CreatedUser(
        user_id=json_response.get('user_id')
    )



import datetime

import jwt

from propelauth_py.errors import UnauthorizedException
from propelauth_py.user import _to_user
import logging

OPTIONS = {
    "verify_signature": True,
    "verify_exp": True,
    "verify_iat": True,
    "verify_iss": True,
    "verify_aud": False,
    "require": ["exp", "iat", "iss"],
}


def _validate_access_token_and_get_user(access_token, token_verification_metadata):
    try:
        decoded_token = jwt.decode(
            access_token,
            token_verification_metadata.verifier_key,
            options=OPTIONS,
            issuer=token_verification_metadata.issuer,
            algorithms=["RS256"],
            leeway=datetime.timedelta(seconds=60),
        )
        return _to_user(decoded_token)
    except UnauthorizedException as e:
        raise e
    except Exception:
        logging.exception("An error occurred while validating the access token")
        raise UnauthorizedException.invalid_access_token()



from collections import namedtuple
from enum import Enum
from uuid import UUID

BACKEND_API_BASE_URL = "https://propelauth-api.com"

TokenVerificationMetadata = namedtuple(
    "TokenVerificationMetadata", ["verifier_key", "issuer"]
)

from requests.auth import AuthBase


class OrgQueryOrderBy(str, Enum):
    CREATED_AT_ASC = "CREATED_AT_ASC"
    CREATED_AT_DESC = "CREATED_AT_DESC"
    NAME = "NAME"


class UserQueryOrderBy(str, Enum):
    CREATED_AT_ASC = "CREATED_AT_ASC"
    CREATED_AT_DESC = "CREATED_AT_DESC"
    LAST_ACTIVE_AT_ASC = "LAST_ACTIVE_AT_ASC"
    LAST_ACTIVE_AT_DESC = "LAST_ACTIVE_AT_DESC"
    EMAIL = "EMAIL"
    USERNAME = "USERNAME"


class _ApiKeyAuth(AuthBase):
    """Attaches API Key Authentication to the given Request object."""

    def __init__(self, integration_api_key):
        self.integration_api_key = integration_api_key

    def __call__(self, r):
        r.headers["Authorization"] = "Bearer " + self.integration_api_key
        return r


def remove_bearer_if_exists(token: str) -> str:
    if not token:
        return token
    elif token.lower().startswith("bearer "):
        return token[7:]
    else:
        return token


def _format_params(params):
    return {
        key: _format_param(value) for key, value in params.items() if value is not None
    }


def _format_param(param):
    if type(param) == bool:
        if param:
            return "true"
        else:
            return "false"
    else:
        return param


def _is_valid_id(identifier):
    try:
        uuid_obj = UUID(identifier, version=4)
        return str(uuid_obj) == identifier
    except ValueError:
        return False


def _is_valid_hex(identifier):
    try:
        int(identifier, 16)
        return True
    except ValueError:
        return False

def _auth_hostname_header(auth_hostname: str) -> dict[str, str]:
    return {"X-Propelauth-url": auth_hostname}


from dataclasses import dataclass
from typing import List

@dataclass
class CustomRoleMapping:
    custom_role_mapping_name: str
    num_orgs_subscribed: int

@dataclass
class CustomRoleMappings:
    custom_role_mappings: List[CustomRoleMapping]


from typing import Optional
import requests
from propelauth_py.api import _ApiKeyAuth, TokenVerificationMetadata, _auth_hostname_header, BACKEND_API_BASE_URL
from propelauth_py.errors import RateLimitedException

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/v1/token_verification_metadata"


####################
#       GET        #
####################
def _fetch_token_verification_metadata(
    auth_hostname: str,
    integration_api_key: str,
    token_verification_metadata: Optional[TokenVerificationMetadata],
):
    if token_verification_metadata is not None:
        return token_verification_metadata

    token_verification_metadata_url = ENDPOINT_URL

    response = requests.get(
        token_verification_metadata_url,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise ValueError("Bad request")
    elif response.status_code == 404:
        raise ValueError("auth_url is incorrect")
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching token verification metadata")

    json_response = response.json()
    return TokenVerificationMetadata(
        verifier_key=json_response.get("verifier_key_pem"),
        issuer="https://" + auth_hostname,
    )



import requests

from propelauth_py.api import _ApiKeyAuth, _auth_hostname_header, BACKEND_API_BASE_URL
from propelauth_py.errors import BadRequestException, RateLimitedException

ENDPOINT_URL = f"{BACKEND_API_BASE_URL}/api/backend/v1/magic_link"

class CreateMagicLinkResponse:
    def __init__(
        self,
        url: str,
    ):
        self.url = url

    def __repr__(self): 
        return (
            f"CreateMagicLinkResponse(url={self.url}"
        )
    def __eq__(self, other):
        return isinstance(other, CreateMagicLinkResponse)


####################
#       POST       #
####################
def _create_magic_link(
    auth_hostname,
    integration_api_key,
    email,
    redirect_to_url=None,
    expires_in_hours=None,
    create_new_user_if_one_doesnt_exist=None,
    user_signup_query_parameters=None,
) -> CreateMagicLinkResponse:
    url = ENDPOINT_URL
    json = {"email": email}
    if redirect_to_url is not None:
        json["redirect_to_url"] = redirect_to_url
    if expires_in_hours is not None:
        json["expires_in_hours"] = expires_in_hours
    if user_signup_query_parameters is not None:
        json["user_signup_query_parameters"] = user_signup_query_parameters
    if create_new_user_if_one_doesnt_exist is not None:
        json[
            "create_new_user_if_one_doesnt_exist"
        ] = create_new_user_if_one_doesnt_exist

    response = requests.post(
        url,
        json=json,
        auth=_ApiKeyAuth(integration_api_key),
        headers=_auth_hostname_header(auth_hostname),
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 429:
        raise RateLimitedException(response.text)
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when creating magic link")

    json_response = response.json()
    return CreateMagicLinkResponse(
        url=json_response.get('url')
    )



from urllib.parse import urlparse


def _validate_and_extract_auth_hostname(auth_url):
    parsed_url = urlparse(auth_url)
    
    if parsed_url.netloc == "":
        raise ValueError("Invalid URL")
    else:
        return parsed_url.netloc
